import sys
import os
from typing import Literal
from pathlib import Path
import pytest
from util import TEST_DIR, template_path_prefix, gen_option_output_map, check_output
from peakrdl.main import main as peakrdl_main


def run_mnrs_exporter(output_dir: Path,
                      gen_option: Literal["addrmap", "regfile", "firmware", "module_header", "module_source"],
                      template_path: str,
                      overwrite: bool = False
                      ):
    arg_list = ["peakrdl", "mnrs", f"--{gen_option}", str(TEST_DIR/"rdl"/"test_addrmap.rdl"),
                "-o", str(output_dir), "--template", template_path]
    if overwrite:
        arg_list.append("--overwrite")
    sys.argv = arg_list
    peakrdl_main()


class TestGeneration():
    def test_addrmap_gen(self, tmp_path: Path):
        run_mnrs_exporter(tmp_path, "addrmap",
                          f"{template_path_prefix}/TestAddrMap.jinja")
        check_output(tmp_path, "addrmap")

    def test_regfile_gen(self, tmp_path: Path):
        run_mnrs_exporter(tmp_path, "regfile",
                          f"{template_path_prefix}/TestRegfile.jinja")
        check_output(tmp_path, "regfile")

    def test_firmware_gen(self, tmp_path: Path):
        run_mnrs_exporter(tmp_path, "firmware",
                          f"{template_path_prefix}/TestRegfile.jinja")
        check_output(tmp_path, "firmware")

    @pytest.mark.filterwarnings('ignore: Using "module_headers" and "template" at the same time')
    def test_module_header_gen(self, tmp_path: Path):
        run_mnrs_exporter(tmp_path, "module_header",
                          f"{template_path_prefix}/TestRegfile.jinja")
        check_output(tmp_path, "module_header")

    @pytest.mark.filterwarnings('ignore: Using "module_sources" and "template" at the same time')
    def test_module_source_gen(self, tmp_path: Path):
        run_mnrs_exporter(tmp_path, "module_source",
                          f"{template_path_prefix}/TestRegfile.jinja")
        check_output(tmp_path, "module_source")


class TestInput():
    def test_invalid_input(self):
        sys.argv = ["peakrdl", "mnrs", "--addrmap",
                    str(TEST_DIR/"rdl"/"invalid_addrmap.rdl"),
                    "-o", "remove_test_dir_should_not_exist"]
        with pytest.raises(ValueError):
            peakrdl_main()

    def test_realtive_path_template(self, tmp_path: Path):
        cur_dir = os.getcwd()
        os.chdir(f"{TEST_DIR.parent}")
        run_mnrs_exporter(tmp_path, "addrmap", "test/TestAddrMap.jinja")
        check_output(tmp_path, "addrmap")
        os.chdir(cur_dir)

    def test_absolute_path_template(self, tmp_path: Path):
        run_mnrs_exporter(tmp_path, "addrmap", str(
            Path(__file__).absolute().parent/'TestAddrMap.jinja'))
        check_output(tmp_path, "addrmap")


class TestOverwrite():
    def test_overwrite(self, tmp_path: Path):
        testfile = f"{tmp_path}/testing_addrmap.h"
        run_mnrs_exporter(tmp_path, "addrmap",
                          f"{template_path_prefix}/TestAddrMap.jinja")

        mtime = os.path.getmtime(testfile)
        run_mnrs_exporter(tmp_path, "addrmap",
                          f"{template_path_prefix}/TestAddrMap.jinja")
        assert mtime == os.path.getmtime(testfile)

        run_mnrs_exporter(tmp_path, "addrmap", f"{template_path_prefix}/TestAddrMap.jinja",
                          overwrite=True)
        assert mtime != os.path.getmtime(testfile)

    @pytest.mark.filterwarnings('ignore: Using "module_headers" and "template"')
    def test_overwrite_protection(self, tmp_path: Path):
        testfiles = [
            f"{tmp_path}/{file}" for file in gen_option_output_map["module_header"]]
        run_mnrs_exporter(tmp_path, "module_header",
                          f"{template_path_prefix}/TestRegfile.jinja")

        mtimes = [os.path.getmtime(testfile) for testfile in testfiles]
        run_mnrs_exporter(tmp_path, "module_header", f"{template_path_prefix}/TestRegfile.jinja",
                          overwrite=True)
        assert mtimes == [os.path.getmtime(testfile) for testfile in testfiles]
